/*
 * BufferSelector.java
 * Copyright (c) 1999, 2000 Chris Laird, Andre Kaplan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


import org.gjt.sp.jedit.Buffer;
import org.gjt.sp.jedit.EBComponent;
import org.gjt.sp.jedit.EBMessage;
import org.gjt.sp.jedit.EditAction;
import org.gjt.sp.jedit.EditBus;
import org.gjt.sp.jedit.EditPane;
import org.gjt.sp.jedit.jEdit;
import org.gjt.sp.jedit.View;
import org.gjt.sp.jedit.msg.BufferUpdate;
import org.gjt.sp.jedit.msg.EditPaneUpdate;

import org.gjt.sp.util.Log;

import java.util.Hashtable;
import java.util.Enumeration;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Insets;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;

import javax.swing.*;


/**
 * A <CODE>BufferSelector</CODE> is a container which holds toggle buttons which
 * allow you to select a buffer. It has the same function as the Buffers menu.
 * Each instance of <CODE>BufferSelector</CODE> is attached to a View.
 * @author Andre Kaplan
 * @author Chris Laird
 * @version $Id: BufferSelector.java,v 1.8 2000/12/31 16:19:43 akaplan Exp $
**/
public class BufferSelector
    extends JPanel
    implements EBComponent
{
    /**
     * Records which <CODE>BufferSelector</CODE> is attached
     * to which <CODE>View</CODE>.
    **/
    private static Hashtable bufferSelectors = new Hashtable();

    /** The margin that the buffer selector buttons will have */
    private static final Insets MARGIN = new Insets(0, 0, 0, 0);

    /* BufferSelector global properties */
    private static boolean defaultEnabled;
    private static boolean useToolbar;
    private static int minWidth;
    private static int maxWidth;
    private static int hgap;
    private static int vgap;

    /** The View that this BufferSelector is attached to */
    private View view;

    private Docker toolbarDocker;
    private Docker viewDocker;

    // Handlers
    private ComponentListener viewResizeHandler;
    private FocusListener textAreaFocusHandler;
    /** The ActionListener that listens to selector button presses */
    private ActionListener actionHandler;

    // Initialize the BufferSelector properties
    static {
        BufferSelector.getProperties();
    }


    /**
     * Create a new <CODE>BufferSelector</CODE> attached to the specified view.
     * @param view the <CODE>View</CODE> that this BufferSelector will be
     *        attached to
    **/
    private BufferSelector(View view) {
        super();

        this.setBorder(
            BorderFactory.createCompoundBorder(
                BorderFactory.createCompoundBorder(
                    BorderFactory.createEmptyBorder(4,0,0,0),
                    BorderFactory.createLoweredBevelBorder()
                ),
                BorderFactory.createEmptyBorder(2,2,1,2)
            )
        );

        this.toolbarDocker = new ToolBarDocker();
        this.viewDocker    = new ViewDocker();

        this.view = view;
        this.viewResizeHandler    = new ViewResizeHandler();
        this.textAreaFocusHandler = new TextAreaFocusHandler();

        this.actionHandler        = new ActionListener() {
            public void actionPerformed(ActionEvent evt) {
                EditAction.getView((Component) evt.getSource())
                    .setBuffer(jEdit.getBuffer(evt.getActionCommand()));
            }
        };
    }


    // EBComponent implementation
    public void handleMessage(EBMessage message) {
        if (message instanceof BufferUpdate) {
            BufferUpdate bu = (BufferUpdate)message;
            if (    (bu.getWhat() == BufferUpdate.CREATED)
                ||  (bu.getWhat() == BufferUpdate.CLOSED)
                ||  (bu.getWhat() == BufferUpdate.DIRTY_CHANGED)
            ) {
                this.update();
                this.view.invalidate();
                this.view.validate();
                this.view.repaint();
            }
        } else if (message instanceof EditPaneUpdate) {
            EditPaneUpdate epu = (EditPaneUpdate) message;
            View v = ((EditPane) epu.getSource()).getView();
            Log.log(Log.DEBUG, this, "***** EditPaneUpdate " + v);
            if (v == this.view) {
                if (epu.getWhat() == EditPaneUpdate.CREATED) {
                    epu.getEditPane().getTextArea().addFocusListener(
                        this.textAreaFocusHandler
                    );
                    this.view.repaint();
                } else if (epu.getWhat() == EditPaneUpdate.DESTROYED) {
                    epu.getEditPane().getTextArea().removeFocusListener(
                        this.textAreaFocusHandler
                    );
                    this.view.repaint();
                } else if (epu.getWhat() == EditPaneUpdate.BUFFER_CHANGED) {
                    this.bufferChanged();
                }
            }
        }
    }


    /**
     * Update the <CODE>BufferSelector</CODE>.
     * This method is called by <CODE>View</CODE> objects when a buffer is
     * opened, closed or a new buffer is selected (either by one of the
     * selector buttons being pressed, or by one of the item in the Buffers
     * menu being chosen).
    **/
    public void update() {
        //get the list of buffers, and add a button for each one
        Buffer[] bufferList = jEdit.getBuffers();

        int minComponentsByLine = this.view.getSize().width / maxWidth;
        if (minComponentsByLine == 0) { minComponentsByLine = 1; }
        int maxComponentsByLine = this.view.getSize().width / minWidth;
        if (maxComponentsByLine == 0) { maxComponentsByLine = 1; }
        /*
        Log.log(Log.DEBUG, this, "view size: " +  this.view.getSize().width);
        Log.log(Log.DEBUG, this, "components: " + bufferList.length);
        Log.log(Log.DEBUG, this, "min components/line: " + minComponentsByLine);
        Log.log(Log.DEBUG, this, "max components/line: " + maxComponentsByLine);
        */
        if (    (this.getLayout() == null)
            ||  !(this.getLayout() instanceof GridLayout)
        ) {
            this.setLayout(new GridLayout(0, minComponentsByLine, hgap, vgap));
        }

        GridLayout gl = (GridLayout)this.getLayout();

        gl.setRows(0);
        if (    bufferList.length >= minComponentsByLine
            &&  bufferList.length <= maxComponentsByLine)
        {
            gl.setColumns(bufferList.length);
        } else if (bufferList.length > maxComponentsByLine) {
            gl.setColumns(maxComponentsByLine);
        } else if (bufferList.length < minComponentsByLine) {
            gl.setColumns(minComponentsByLine);
        }

        gl.setHgap(hgap);
        gl.setVgap(vgap);

        int fontSize = Integer.parseInt(
            jEdit.getProperty("buffer-selector.font-size"));

        //remove all buttons from the selector, removing them as
        //ActionListeners and BufferListeners first
        this.removeAllButtons();
        ButtonGroup group = new ButtonGroup();
        //stores the font to be used for the buttons
        Font buttonFont = null;

        for (int i = 0; i < bufferList.length; i++) {
            Buffer buffer = bufferList[i];

            //use the buffer title as the text on the button
            String name = buffer.getName();
            JToggleButton toggle = new JToggleButton(name);
            toggle.setMargin(MARGIN);
            toggle.setHorizontalAlignment(JToggleButton.LEFT);

            //when the first button is created, read its font
            if (buttonFont == null) {
                Font f = toggle.getFont();
                buttonFont = new Font(f.getName(), Font.PLAIN, fontSize);
            }
            toggle.setFont(buttonFont);

            if (buffer.isReadOnly()) {
                toggle.setForeground(Color.gray);
            }

            //set up the tooltip (the full path to the buffer)
            String tooltip = buffer.getPath();
            toggle.setToolTipText(tooltip);

            //set up the icon
            Icon toggleIcon = buffer.getIcon();
            toggle.setIcon(toggleIcon);

            //if this is the current buffer, make the button selected
            //and use a bold font
            if(view.getBuffer() == buffer) {
                toggle.setSelected(true);
                toggle.setFont(new Font(buttonFont.getName(), Font.BOLD, fontSize));
            }

            group.add(toggle);
            toggle.setActionCommand(buffer.getPath());
            toggle.addActionListener(actionHandler);

            this.add(toggle);
        }
    }


    public void bufferChanged() {
        this.bufferChanged(this.view.getBuffer());
    }


    public void bufferChanged(Buffer buffer) {
        // Log.log(Log.DEBUG, this, "bufferChanged: " + buffer.getName());
        Component[] component = this.getComponents();
        Font buttonFont = null;
        int fontSize = Integer.parseInt(
            jEdit.getProperty("buffer-selector.font-size"));

        for (int i = 0; i < component.length; i++) {
            if (component[i] instanceof AbstractButton) {
                AbstractButton toggle = ((AbstractButton)component[i]);
                if (buttonFont == null) {
                    Font f = toggle.getFont();
                    buttonFont = new Font(f.getName(), Font.PLAIN, fontSize);
                }

                if (toggle.getActionCommand().equals(buffer.getPath())) {
                    toggle.setSelected(true);
                    toggle.setFont(new Font(buttonFont.getName(),
                        Font.BOLD, fontSize));
                } else {
                    toggle.setFont(buttonFont);
                }
            }
        }
    }


    /**
     * Remove all buttons from the <CODE>BufferSelector</CODE>, unregistering
     * them with the <CODE>ActionListener</CODE> first.
    **/
    private void removeAllButtons() {
        // unregister all buttons with the ActionListener
        Component[] component = this.getComponents();
        for (int i = 0; i < component.length; i++) {
            if (component[i] instanceof AbstractButton) {
                ((AbstractButton) component[i]).removeActionListener(
                    this.actionHandler
                );
            }
        }
        // remove all buttons from the BufferSelector
        this.removeAll();
    }


    /**
     * Adds event handlers associated with this Buffer Selector
     * <UL>
     * <LI><EM>View Resize Handler</EM>: updates this when the View is resized
     * <LI><EM>Focus Handler</EM>: added to all view edit panes, updates this
     * when an EditPane gains focus
     * </UL>
    **/
    private void addHandlers() {
        this.view.addComponentListener(this.viewResizeHandler);

        EditPane[] editPanes = this.view.getEditPanes();
        for (int i = 0; i < editPanes.length; i++) {
            editPanes[i].getTextArea().addFocusListener(
                this.textAreaFocusHandler);
        }
    }


    /**
     * Removes event handlers associated with this Buffer Selector
    **/
    private void removeHandlers() {
        this.view.removeComponentListener(this.viewResizeHandler);

        EditPane[] editPanes = this.view.getEditPanes();
        for (int i = 0; i < editPanes.length; i++) {
            editPanes[i].getTextArea().removeFocusListener(
                this.textAreaFocusHandler);
        }
    }


    /**
     * Tells if a <CODE>BufferSelector</CODE> is enabled for a specified
     * View
    **/
    public static boolean isEnabledFor(View view) {
        return (bufferSelectors.get(view) != null);
    }


    /**
     * Adds a <CODE>BufferSelector</CODE> to a specified View
    **/
    public static void addTo(View view) {
        BufferSelector bs = new BufferSelector(view);

        // Add View Resize Handlers and Text Area Handlers
        bs.removeHandlers(); // Necessary for EditPanes focus handlers
        bs.addHandlers();

        // Add the BufferSelector to the EditBus
        EditBus.addToBus(bs);

        // Adds a BufferSelector <-> View association
        bufferSelectors.put(view, bs);

        // Add the BufferSelector to the View
        if (BufferSelector.useToolbar) {
            bs.toolbarDocker.add();
        } else {
            bs.viewDocker.add();
        }

        bs.update();
        view.invalidate();
        view.validate();
    }


    /**
     * Removes a <CODE>BufferSelector</CODE> from a specified View
    **/
    public static void removeFrom(View view) {
        BufferSelector bs = (BufferSelector) bufferSelectors.get(view);
        if (bs != null) {
            Container cp = (Container) view.getContentPane();

            if (bs.viewDocker.isVisible()) {
                bs.viewDocker.remove();
            }

            if (bs.toolbarDocker.isVisible()) {
                bs.toolbarDocker.remove();
            }

            // Remove View Resize Handlers and Text Area Handlers
            bs.removeHandlers();
            // Remove from EditBus
            EditBus.removeFromBus(bs);
            // Removes the BufferSelector <-> View association
            bufferSelectors.remove(view);

            view.invalidate();
            view.validate();
        }
    }


    /**
     * Toggles a <CODE>BufferSelector</CODE> for a specified View
    **/
    public static void toggleFor(View view) {
        if (BufferSelector.isEnabledFor(view)) {
            BufferSelector.removeFrom(view);
        } else {
            BufferSelector.addTo(view);
        }
    }


    public static void propertiesChanged() {
        boolean oldUseToolbar = BufferSelector.useToolbar;
        BufferSelector.getProperties();
        View[] views = jEdit.getViews();

        for (int i = 0; i < views.length; i++) {
            if (BufferSelector.isEnabledFor(views[i])) {
                if (!defaultEnabled) {
                    BufferSelector.removeFrom(views[i]);
                } else {
                    if (oldUseToolbar != BufferSelector.useToolbar) {
                        BufferSelector.removeFrom(views[i]);
                        BufferSelector.addTo(views[i]);
                    }
                }
            } else {
                if (defaultEnabled) {
                    BufferSelector.addTo(views[i]);
                }
            }
        }

        for (Enumeration e = bufferSelectors.elements(); e.hasMoreElements();) {
            BufferSelector bs = (BufferSelector) e.nextElement();
            bs.update();
        }
    }


    private static void getProperties() {
        defaultEnabled = jEdit.getBooleanProperty(
            "buffer-selector.default-enabled", false);
        useToolbar = jEdit.getBooleanProperty(
            "buffer-selector.use-toolbar", false);

        minWidth = Integer.parseInt(jEdit.getProperty(
            "buffer-selector.min-width"));
        maxWidth = Integer.parseInt(jEdit.getProperty(
            "buffer-selector.max-width"));
        hgap = Integer.parseInt(jEdit.getProperty("button.hgap"));
        vgap = Integer.parseInt(jEdit.getProperty("button.vgap"));
    }


    // To keep track of the View being resized
    private class ViewResizeHandler extends ComponentAdapter
    {
        public void componentResized(ComponentEvent evt) {
            BufferSelector.this.update();
            BufferSelector.this.view.invalidate();
            BufferSelector.this.view.validate();
        }
    }


    // BufferSelector acts on the last focused TextArea
    private class TextAreaFocusHandler extends FocusAdapter {
        public void focusGained(FocusEvent evt) {
            // walk up hierarchy, looking for an EditPane
            Component comp = (Component) evt.getSource();
            while(!(comp instanceof EditPane)) {
                if (comp == null) {
                    return;
                }
                comp = comp.getParent();
            }

            BufferSelector.this.bufferChanged(((EditPane) comp).getBuffer());
            BufferSelector.this.view.invalidate();
            BufferSelector.this.view.validate();
        }
    }


    /** Responsible for docking BufferSelector at the right place */
    private interface Docker {
        boolean isVisible();

        void add();
        void remove();
    }


    private class ToolBarDocker implements Docker {
        private JToolBar toolbar = null;


        public boolean isVisible() {
            return (this.toolbar != null);
        }


        /**
         * Adds BufferSelector as a jEdit toolbar
        **/
        public void add() {
            this.toolbar = new JToolBar();
            this.toolbar.setLayout(new BorderLayout());
            this.toolbar.setFloatable(false);

            this.toolbar.add(BorderLayout.CENTER, BufferSelector.this);
            BufferSelector.this.view.addToolBar(this.toolbar);
        }


        /**
         * Removes BufferSelector from the jEdit toolbars
        **/
        public void remove() {
            BufferSelector.this.view.removeToolBar(this.toolbar);
            this.toolbar = null;
        }
    }


    /** ViewDocker works with jEdit 2.7pre2 and (hopefully) above */
    private class ViewDocker implements Docker {
        private Box       box;


        /**
         * true if this BufferSelector was added below the CommandLine
        **/
        public boolean isVisible() {
            return (this.box != null);
        }


        /**
         * Adds BufferSelector below the view edit panes (creating a Box)
        **/
        public void add() {
            Log.log(Log.DEBUG, this, "Adding to view...");

            Container cp = BufferSelector.this.view.getContentPane();

            this.box = new Box(BoxLayout.Y_AXIS);
            this.box.add(BufferSelector.this);

            cp.add(this.box, BorderLayout.SOUTH);

            Log.log(Log.DEBUG, this, "... Added to view");
        }


        /**
         * Removes BufferSelector from below the view edit panes
        **/
        public void remove() {
            Log.log(Log.DEBUG, this, "Removing from view...");

            Container cp = BufferSelector.this.view.getContentPane();
            cp.remove(this.box);

            this.box         = null;

            Log.log(Log.DEBUG, this, "... Removed from view");
        }
    }
}
